/**
 *	File:       	5110.cpp
 *	Version:  		1.0
 *	Date:       	2013
 *	License:		GPL v3
 *	Description:	unbuffered driver for Nokia 3110 / 5110 LCD
 *	Project:		uRADMonitor KIT1, a hackable digital sensor monitoring tool with network interface
 *  
 *	Copyright 2007 by Tony Myatt
 *	Copyright 2013 by Radu Motisan, radu.motisan@gmail.com
 *	Copyright 2016 by Magnasci SRL, www.magnasci.com
 *  
 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 * 	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * 	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <avr/pgmspace.h>

#include "5110.h"


// Alphabet lookup font
// we are actually using 6x8 pixels, but the last are for border
// this means in 84x48 we have 14x6 characters
// the list is ordonated by the ASCII codes 
const uint8_t PROGMEM font5x7 [][CHAR_WIDTH - 1] = {
	{ 0x00, 0x00, 0x00, 0x00, 0x00 },   // sp
    { 0x00, 0x00, 0x2f, 0x00, 0x00 },   // !
    { 0x00, 0x07, 0x00, 0x07, 0x00 },   // "
    { 0x14, 0x7f, 0x14, 0x7f, 0x14 },   // #
    { 0x24, 0x2a, 0x7f, 0x2a, 0x12 },   // $
	{ 0x32, 0x34, 0x08, 0x16, 0x26 },   // %
    { 0x36, 0x49, 0x55, 0x22, 0x50 },   // &
    { 0x00, 0x05, 0x03, 0x00, 0x00 },   // '
    { 0x00, 0x1c, 0x22, 0x41, 0x00 },   // (
    { 0x00, 0x41, 0x22, 0x1c, 0x00 },   // )
    { 0x14, 0x08, 0x3E, 0x08, 0x14 },   // *
    { 0x08, 0x08, 0x3E, 0x08, 0x08 },   // +
    { 0x00, 0x00, 0x50, 0x30, 0x00 },   // ,
    { 0x10, 0x10, 0x10, 0x10, 0x10 },   // -
    { 0x00, 0x60, 0x60, 0x00, 0x00 },   // .
    { 0x20, 0x10, 0x08, 0x04, 0x02 },   // /
    { 0x3E, 0x51, 0x49, 0x45, 0x3E },   // 0
    { 0x00, 0x42, 0x7F, 0x40, 0x00 },   // 1
    { 0x42, 0x61, 0x51, 0x49, 0x46 },   // 2
    { 0x21, 0x41, 0x45, 0x4B, 0x31 },   // 3
    { 0x18, 0x14, 0x12, 0x7F, 0x10 },   // 4
    { 0x27, 0x45, 0x45, 0x45, 0x39 },   // 5
    { 0x3C, 0x4A, 0x49, 0x49, 0x30 },   // 6
    { 0x01, 0x71, 0x09, 0x05, 0x03 },   // 7
    { 0x36, 0x49, 0x49, 0x49, 0x36 },   // 8
    { 0x06, 0x49, 0x49, 0x29, 0x1E },   // 9
    { 0x00, 0x36, 0x36, 0x00, 0x00 },   // :
    { 0x00, 0x56, 0x36, 0x00, 0x00 },   // ;
    { 0x08, 0x14, 0x22, 0x41, 0x00 },   // <
    { 0x14, 0x14, 0x14, 0x14, 0x14 },   // =
    { 0x00, 0x41, 0x22, 0x14, 0x08 },   // >
    { 0x02, 0x01, 0x51, 0x09, 0x06 },   // ?
    { 0x32, 0x49, 0x59, 0x51, 0x3E },   // @
    { 0x7E, 0x11, 0x11, 0x11, 0x7E },   // A
    { 0x7F, 0x49, 0x49, 0x49, 0x36 },   // B
    { 0x3E, 0x41, 0x41, 0x41, 0x22 },   // C
    { 0x7F, 0x41, 0x41, 0x22, 0x1C },   // D
    { 0x7F, 0x49, 0x49, 0x49, 0x41 },   // E
    { 0x7F, 0x09, 0x09, 0x09, 0x01 },   // F
    { 0x3E, 0x41, 0x49, 0x49, 0x7A },   // G
    { 0x7F, 0x08, 0x08, 0x08, 0x7F },   // H
    { 0x00, 0x41, 0x7F, 0x41, 0x00 },   // I
    { 0x20, 0x40, 0x41, 0x3F, 0x01 },   // J
    { 0x7F, 0x08, 0x14, 0x22, 0x41 },   // K
    { 0x7F, 0x40, 0x40, 0x40, 0x40 },   // L
    { 0x7F, 0x02, 0x0C, 0x02, 0x7F },   // M
    { 0x7F, 0x04, 0x08, 0x10, 0x7F },   // N
    { 0x3E, 0x41, 0x41, 0x41, 0x3E },   // O
    { 0x7F, 0x09, 0x09, 0x09, 0x06 },   // P
    { 0x3E, 0x41, 0x51, 0x21, 0x5E },   // Q
    { 0x7F, 0x09, 0x19, 0x29, 0x46 },   // R
    { 0x46, 0x49, 0x49, 0x49, 0x31 },   // S
    { 0x01, 0x01, 0x7F, 0x01, 0x01 },   // T
    { 0x3F, 0x40, 0x40, 0x3F },   		// U
    { 0x1F, 0x20, 0x40, 0x20, 0x1F },   // V
    { 0x3F, 0x40, 0x38, 0x40, 0x3F },   // W
    { 0x63, 0x14, 0x08, 0x14, 0x63 },   // X
    { 0x07, 0x08, 0x70, 0x08, 0x07 },   // Y
    { 0x61, 0x51, 0x49, 0x45, 0x43 },   // Z
    { 0x00, 0x7F, 0x41, 0x41, 0x00 },   // [
    { 0x55, 0x2A, 0x55, 0x2A, 0x55 },   // 55 /
    { 0x00, 0x41, 0x41, 0x7F, 0x00 },   // ]
    { 0x04, 0x02, 0x01, 0x02, 0x04 },   // ^
    { 0x40, 0x40, 0x40, 0x40, 0x40 },   // _
    { 0x00, 0x01, 0x02, 0x04, 0x00 },   // '
    { 0x20, 0x54, 0x54, 0x54, 0x78 },   // a
    { 0x7F, 0x48, 0x44, 0x44, 0x38 },   // b
    { 0x38, 0x44, 0x44, 0x44, 0x20 },   // c
    { 0x38, 0x44, 0x44, 0x48, 0x7F },   // d
    { 0x38, 0x54, 0x54, 0x54, 0x18 },   // e
    { 0x08, 0x7E, 0x09, 0x01, 0x02 },   // f
    { 0x0C, 0x52, 0x52, 0x52, 0x3E },   // g
    { 0x7F, 0x08, 0x04, 0x04, 0x78 },   // h
    { 0x00, 0x44, 0x7D, 0x40, 0x00 },   // i
    { 0x20, 0x40, 0x44, 0x3D, 0x00 },   // j
    { 0x7F, 0x10, 0x28, 0x44, 0x00 },   // k
    { 0x00, 0x41, 0x7F, 0x40, 0x00 },   // l
    { 0x7C, 0x04, 0x18, 0x04, 0x78 },   // m
    { 0x7C, 0x08, 0x04, 0x04, 0x78 },   // n
    { 0x38, 0x44, 0x44, 0x44, 0x38 },   // o
    { 0x7C, 0x14, 0x14, 0x14, 0x08 },   // p
    { 0x08, 0x14, 0x14, 0x18, 0x7C },   // q
    { 0x7C, 0x08, 0x04, 0x04, 0x08 },   // r
    { 0x48, 0x54, 0x54, 0x54, 0x20 },   // s
    { 0x04, 0x3F, 0x44, 0x40, 0x20 },   // t
    { 0x3C, 0x40, 0x40, 0x20, 0x7C },   // u
    { 0x1C, 0x20, 0x40, 0x20, 0x1C },   // v
    { 0x3C, 0x40, 0x30, 0x40, 0x3C },   // w
    { 0x44, 0x28, 0x10, 0x28, 0x44 },   // x
    { 0x0C, 0x50, 0x50, 0x50, 0x3C },   // y
    { 0x44, 0x64, 0x54, 0x4C, 0x44 },   // z 0x7A
    { 0x00, 0x7F, 0x3E, 0x1C, 0x08 },   // > Filled {
	{ 0x08, 0x1C, 0x3E, 0x7F, 0x00 }, 	 // < Filled |
	{ 0x08, 0x7C, 0x7E, 0x7C, 0x08 },   // Arrow up }
	{ 0x10, 0x3E, 0x7E, 0x3E, 0x10 },   // Arrow down ~	
	{ 0x3E, 0x3E, 0x3E, 0x3E, 0x3E },   // Stop 0x7F
	{ 0x00, 0x7F, 0x3E, 0x1C, 0x08 },   // Play 0x80
	/*{ 0x7F, 0x7F, 0x7F, 0x7F, 0x7F },	// bat 1 0x81
	{ 0x77, 0x7B, 0x41, 0x7B, 0x77 },	// bat 2 0x82
	{ 0x7F, 0x7F, 0x7F, 0x1C, 0x1C },	// bat 3 0x83*/
	{ 0x7F, 0x41, 0x5D, 0x55, 0x55 },	// bat 1 0x81
	{ 0x49, 0x49, 0x49, 0x5D, 0x49 },	// bat 1 0x82 //charging
	{ 0x41, 0x41, 0x41, 0x7F, 0x1C },	// bat 1 0x83
	{ 0x7F, 0x41, 0x5D, 0x5D, 0x5D },	// bat 2 0x84
	{ 0x5D, 0x5D, 0x5D, 0x5D, 0x5D },	// bat 2 0x85 //full
	{ 0x5D, 0x5D, 0x41, 0x7F, 0x1C },	// bat 2 0x86
	{ 0x7F, 0x41, 0x5D, 0x5D, 0x5D },	// bat 3 0x87
	{ 0x5D, 0x5D, 0x59, 0x51, 0x41 },	// bat 3 0x88 //half
	{ 0x41, 0x41, 0x41, 0x7F, 0x1C },	// bat 3 0x89
	{ 0x7F, 0x41, 0x5D, 0x59, 0x51 },	// bat 4 0x8A
	{ 0x41, 0x41, 0x41, 0x41, 0x41 },	// bat 4 0x8B //empty
	{ 0x41, 0x41, 0x41, 0x7F, 0x1C },	// bat 4 0x8C
	{ 0x40, 0x3C, 0x10, 0x1C, 0x00 },	// micro 0x8D
	{ 0x04, 0x26, 0x38, 0x26, 0x04 },	// radsign 0x8E
	{ 0x67, 0x15, 0x60, 0x17, 0x63 },	// CPM sign 0x8F
	{ 0x7C, 0x7C, 0x0C, 0x0C, 0x0C },	//  sign 0x90
	{ 0x0C, 0x0C, 0x0C, 0x7C, 0x7C },	//  sign 0x91
	{ 0x7F, 0x7F, 0x00, 0x00, 0x00 },	//  sign 0x92
	{ 0x00, 0x00, 0x00, 0x7F, 0x7F },	//  sign 0x93
	{ 0x0C, 0x0C, 0x0C, 0x0C, 0x0C },	//  sign 0x94
	{ 0x0F, 0x0F, 0x0C, 0x0C, 0x0C },	//  sign 0x95
	{ 0x0C, 0x0C, 0x0C, 0x0F, 0x0F },	//  sign 0x96
	{ 0x1C, 0x1C, 0x3E, 0x7F, 0x00 },   //  speaker on 0x97
	{ 0x1A, 0x14, 0x2E, 0x5F, 0x20 },   //  speaker off 0x98
	{ 0x07, 0x75, 0x5D, 0x57, 0x70 },   //  lan on 0x99
	{ 0x05, 0x72, 0x5D, 0x50, 0x70 },   //  lan off 0x9A
	{ 0x00, 0x07, 0x05, 0x07, 0x00 },	// degrees 0x9B
	{ 0x1F, 0x05, 0x72, 0x28, 0x70 }	// Pascals 0x9C
};


LCD_5110::LCD_5110(DigitalPin *rst, DigitalPin *ce, DigitalPin *dc, DigitalPin *data, DigitalPin *clk, DigitalPin *backlight) {
	lcdCacheIdx = 0;
	m_rst = rst;
	m_ce = ce;
	m_dc = dc;
	m_data = data;
	m_clk = clk;
	m_backlight = backlight;
}

/* Performs IO & LCD controller initialization */
void LCD_5110::init() {
	// Pull-up on reset pin
	*m_rst = 1;
	
	// Wait after VCC high for reset (max 30ms)
    _delay_ms(15);
    
    // Toggle display reset pin
    *m_rst = 0;
    _delay_ms(64);
    *m_rst = 1;

    // Disable LCD controller
    *m_ce = 0;

    send(LCD_EXTENDED_COMMANDS, LCD_CMD);  
    send(LCD_VOP, LCD_CMD);  
    send(LCD_TEMP, LCD_CMD);  
    send(LCD_BIAS_MODE, LCD_CMD);  
    send(LCD_HOR_ADDRESSING, LCD_CMD);  
    send(LCD_NORMAL_MODE, LCD_CMD);  
    
    // Clear lcd
    clear();

	// Set display contrast. Note: No change is visible at ambient temperature 
	int contrast = 0x40;
	send(LCD_EXTENDED_COMMANDS, LCD_CMD);	// LCD Extended Commands
	send(0x80 | contrast, LCD_CMD);			// Set LCD Vop(Contrast)
	send(LCD_HOR_ADDRESSING, LCD_CMD);		// LCD std cmds, hori addr mode
}



// Clears the display 
void LCD_5110::clear(void) {
	lcdCacheIdx = 0;
	base_addr(lcdCacheIdx);
	// Set the entire cache to zero and write 0s to lcd
	
	int size =  ((LCD_WIDTH * LCD_HEIGHT) / CHAR_HEIGHT);
	
    for(int i=0;i<size;i++)  send(0, LCD_DATA);
    
}

// Clears an area on a line 
void LCD_5110::clear_area(uint8_t line, uint8_t startX, uint8_t endX) {
    // Start and end positions of line
    int start = line * LCD_WIDTH + startX;
    int end = line * LCD_WIDTH + endX;
	
	base_addr(start);
    
    // Clear all data in range from cache
    for(uint16_t i = start; i < end; i++) send(0, LCD_DATA);
}

// Clears an entire text block. (rows of 8 pixels on the lcd) 
void LCD_5110::clear_line(uint8_t line) {
    clear_area(line, 0, LCD_WIDTH);
}

// Sets cursor location to xy location corresponding to basic font size 
void LCD_5110::goto_xy(uint8_t x, uint8_t y) {
    lcdCacheIdx = x*CHAR_WIDTH + y * LCD_WIDTH;
}

// Sets cursor location to exact xy pixel location on the lcd 
void LCD_5110::goto_xy_exact(uint8_t x, uint8_t y) {
    lcdCacheIdx = x + y * LCD_WIDTH;
}

// Sends data to display controller : a line of 8 pixels, as defined by data
void LCD_5110::send(uint8_t byte, Type cd) {
	// Enable display controller (active low)
	*m_ce = 0;
    // Either command or data
    *m_dc =  (cd == LCD_DATA);
	for(uint8_t i=0;i<8;i++) {
		// Set the DATA pin value:8 bits in a row
		*m_data = (byte >> (7-i)) & 0x01;
		// Toggle the clock after each data bit
		*m_clk = 1;
		*m_clk = 0;
	}
	// Disable display controller
	*m_ce = 1;
}

// Displays a character at current cursor location or moves to new line if enter char is given
void LCD_5110::send(char chr) {
	if (chr == '\n') {
		// move to next line and first column
		lcdCacheIdx = lcdCacheIdx + LCD_WIDTH - lcdCacheIdx % LCD_WIDTH; //next line and x=0
	} else {
		base_addr(lcdCacheIdx);

		// 5 pixel wide characters and add space
		for(uint8_t i=0; i < CHAR_WIDTH - 1; i++)
			send(pgm_read_byte(&font5x7[chr-32][i]) << 1, LCD_DATA);

		send(0, LCD_DATA); // right empty separator line of 8 vertical pixels

		lcdCacheIdx += CHAR_WIDTH;
	}
}

// Displays null terminated string at current cursor location and increment cursor location 
void LCD_5110::send(char *str) {
    while(*str)  send(*str++);
}

void LCD_5110::send(const char *str) {
    while(pgm_read_byte(str)) send(pgm_read_byte(str++));
}

// displays a formated string, similar to printf
void LCD_5110::send(char *buffer, uint16_t len, const char *szFormat, ...) {
	va_list pArgs;
	va_start(pArgs, szFormat);
	vsnprintf_P(buffer, len, szFormat, pArgs);
	va_end(pArgs);
	send(buffer);
}

// Set the base address of the lcd
void LCD_5110::base_addr(uint16_t addr) {
	send(LCD_SETXADDR |(addr % LCD_WIDTH), LCD_CMD);
	send(LCD_SETYADDR |(addr / LCD_WIDTH), LCD_CMD);
}



void LCD_5110::col(char chr) {
	base_addr(lcdCacheIdx);
	send(chr, LCD_DATA);
	lcdCacheIdx++;
}

// It goes back the cursor on LCD for a single step
void LCD_5110::pixelBack(void)  {
	lcdCacheIdx--;
}

// Prints on LCD a hex based picture, A hex picture can be produced from the "LCDAssistant.exe" windows based software.  
void LCD_5110::printPictureOnLCD ( const uint8_t *data) {
	uint16_t pixel_cols = LCD_WIDTH * (LCD_HEIGHT / CHAR_HEIGHT); //<6> means 6 lines on LCD.
 	goto_xy(0, 0);
	for(int i=0;i<pixel_cols;i++) col(pgm_read_byte(data++));
}

  
// the most basic function, set a single pixel
void LCD_5110::drawPixel(uint8_t  x, uint8_t  y, int color) {
	uint16_t pixel_addr = x + y/CHAR_HEIGHT * LCD_WIDTH;
	send(LCD_SETXADDR |(pixel_addr % LCD_WIDTH), LCD_CMD);
	send(LCD_SETYADDR |(pixel_addr / LCD_WIDTH), LCD_CMD);
	if (color)
		send(1<<(y % CHAR_HEIGHT), LCD_DATA); //"black pixel"
	else
		send(0, LCD_DATA); //"remove black pixel"
}

void LCD_5110::setBacklight(bool mode) {
	*m_backlight = mode;
}
